﻿using System.Collections.Generic;
using System.Threading;

namespace ComputeFileHashes
{
    /// <summary>
    /// Implements a generic Queue that blocks callers of Dequeue when empty.
    /// </summary>
    /// <typeparam name="T">Type parameter</typeparam>
    class BlockingQueue<T> : Queue<T>
    {
        // Locking object
        private readonly object _lock = new object();

        /// <summary>
        /// Dequeues an item if present; blocks the calling thread otherwise.
        /// </summary>
        /// <returns>Dequeued item.</returns>
        public T BlockingDequeue()
        {
            lock (_lock)
            {
                // Block if queue is empty
                if (0 == Count)
                {
                    Monitor.Wait(_lock);
                }

                // Dequeue and return item
                return Dequeue();
            }
        }

        /// <summary>
        /// Enqueues the specified item and unblocks a calling thread if necessary.
        /// </summary>
        /// <param name="item">Item to enqueue.</param>
        public void UnblockingEnqueue(T item)
        {
            lock (_lock)
            {
                // Enqueue the item and unblock a waiting thread.
                Enqueue(item);
                Monitor.Pulse(_lock);
            }
        }
    }
}
