﻿using System;
using System.Collections.Generic;
using System.Diagnostics.CodeAnalysis;
using System.Globalization;
using System.IO;
using System.Security.Cryptography;
using System.Text;
using System.Threading;
using Delay;

namespace ComputeFileHashes
{
    /// <summary>
    /// Computes multiple hashes for the specified file(s).
    /// </summary>
    [SuppressMessage("Microsoft.Design", "CA1063:ImplementIDisposableCorrectly", Justification = "No need to clean up unmanaged resources")]
    public class ComputeFileHashesCore : IDisposable
    {
        /// <summary>
        /// Version number of the implementation.
        /// </summary>
        public const string Version = "2010-11-30";

        /// <summary>
        /// Collection of supported HashAlgorithms.
        /// </summary>
        public Dictionary<string, HashAlgorithm> Hashes { get; private set; }

        // 1MB buffer for reading from the disk
        private byte[] _buffer;
        // Same-sized buffer for swapping with
        private byte[] _nextBuffer;
        // Count of valid bytes in _buffer
        private int _bufferCount;
        // True iff done
        private bool _done;
        // Synchronizes the worker thread readiness
        private readonly WaitingRoom _waitingForReadiness;
        // Synchronizes worker thread completion
        private readonly WaitingRoom _waitingForCompletion;

        /// <summary>
        /// Initializes a new instance.
        /// </summary>
        public ComputeFileHashesCore()
        {
            // Initialize variables
            Hashes = new Dictionary<string, HashAlgorithm>
            {
                { "CRC32", new CRC32() },
#if SILVERLIGHT
                { "MD5", new MD5Managed() },
                { "SHA1", new SHA1Managed() },
                { "SHA256", new SHA256Managed() },
#else
                { "MD5", MD5.Create() },
                { "SHA1", SHA1.Create() },
                { "SHA256", SHA256.Create() },
                { "SHA384", SHA384.Create() },
                { "SHA512", SHA512.Create() },
                { "RIPEMD160", RIPEMD160.Create() },
#endif
            };
            _waitingForReadiness = new WaitingRoom(Hashes.Count);
            _waitingForCompletion = new WaitingRoom(Hashes.Count);
            _buffer = new byte[1024 * 1024];
            _nextBuffer = new byte[_buffer.Length];

            // Check hash algorithms for necessary functionality
            foreach (var hash in Hashes)
            {
                var hashAlgorithm = hash.Value;
                if (!hashAlgorithm.CanReuseTransform || !hashAlgorithm.CanTransformMultipleBlocks)
                {
                    throw new NotImplementedException("One or more hash algorithms do not support reuse and/or multiple blocks.");
                }
                (new Thread(HashThreadFunction)).Start(hashAlgorithm);
            }
        }

        /// <summary>
        /// Implement IDisposable.Dispose to clean up.
        /// </summary>
        [SuppressMessage("Microsoft.Design", "CA1063:ImplementIDisposableCorrectly", Justification = "No need to clean up unmanaged resources")]
        [SuppressMessage("Microsoft.Usage", "CA1816:CallGCSuppressFinalizeCorrectly", Justification = "Does not implement a finalizer")]
        public void Dispose()
        {
            // Out of files; tell worker threads to exit
            _done = true;
            _waitingForReadiness.Wait();
            _waitingForReadiness.Release();
        }

        /// <summary>
        /// Processes a single file.
        /// </summary>
        /// <param name="fileInfo">File to process.</param>
        /// <param name="progressCallback">Method to call with progress updates.</param>
        public void ProcessFile(FileInfo fileInfo, ProgressHandler progressCallback)
        {
            // Validate arguments
            if (null == fileInfo)
            {
                throw new ArgumentNullException("fileInfo");
            }
            if (null == progressCallback)
            {
                throw new ArgumentNullException("progressCallback");
            }

            // Initialize variables
            long bytesProcessed = 0;

            // Open and process file
            using (var stream = fileInfo.OpenRead())
            {
                // Read initial block
                _bufferCount = stream.Read(_buffer, 0, _buffer.Length);

                // Wait for worker threads to become ready and start them
                _waitingForReadiness.Wait();
                _waitingForReadiness.Release();

                do
                {
                    // Update status
                    bytesProcessed += _bufferCount;
                    var proceed = progressCallback((0 < fileInfo.Length ? ((double)bytesProcessed / fileInfo.Length) : 1) * 100);

                    // Read next block
                    var nextBufferCount = proceed ?
                        stream.Read(_nextBuffer, 0, _nextBuffer.Length) :
                        0;

                    // Wait for worker threads to finish and swap buffers
                    _waitingForCompletion.Wait();
                    var tempBuffer = _buffer;
                    _buffer = _nextBuffer;
                    _nextBuffer = tempBuffer;
                    _bufferCount = nextBufferCount;

                    // Let worker threads continue with next block
                    _waitingForCompletion.Release();

                    // Until EOF
                } while (0 < _bufferCount);

                // Wait for worker threads to finish last block
                _waitingForCompletion.Wait();

                // Let worker threads get ready for next file
                _waitingForCompletion.Release();
            }
        }

        /// <summary>
        /// Implements code for worker threads.
        /// </summary>
        /// <param name="param">HashAlgorithm to use.</param>
        private void HashThreadFunction(object param)
        {
            // Initialize variables
            var hashAlgorithm = (HashAlgorithm)param;

            try
            {
                // Until time to quit
                while (true)
                {
                    // Indicate readiness
                    _waitingForReadiness.Arrive();

                    // Exit if time to quit
                    if (_done)
                    {
                        break;
                    }

                    // Initialize hash algorithm
                    hashAlgorithm.Initialize();

                    do
                    {
                        // Transform current block and indicate completion
                        hashAlgorithm.TransformBlock(_buffer, 0, _bufferCount, null, 0);
                        _waitingForCompletion.Arrive();

                        // Until EOF
                    } while (0 < _bufferCount);

                    // Transform final block and indicate completion
                    hashAlgorithm.TransformFinalBlock(_buffer, 0, _bufferCount);
                    _waitingForCompletion.Arrive();
                }
            }
            catch(ThreadAbortException)
            {
                // Unexpected abort (ex: browser closed); signal readiness for owner thread in Dispose
                _waitingForReadiness.Arrive();
            }
        }

        /// <summary>
        /// Convert a byte array to a string of hex values.
        /// </summary>
        /// <param name="bytes">Byte array.</param>
        /// <returns>String of hex values.</returns>
        public static string ConvertValueToString(byte[] value)
        {
            // Validate argument
            if (null == value)
            {
                throw new ArgumentNullException("value");
            }

            // Create a properly-sized buffer
            var sb = new StringBuilder(value.Length * 2);

            // Add each byte's hex value
            foreach(var b in value)
            {
                sb.AppendFormat(CultureInfo.CurrentCulture, "{0:X2}", b);
            }

            // Return the result
            return sb.ToString();
        }
    }

    /// <summary>
    /// Delegate for receiving progress updates.
    /// </summary>
    /// <param name="progress">Progress as a percent.</param>
    /// <returns>True to continue; false to abort.</returns>
    public delegate bool ProgressHandler(double progress);
}
