﻿using System;
using System.Collections.ObjectModel;
using System.Diagnostics.CodeAnalysis;
using System.IO;
using System.Threading;
using System.Windows.Controls;
using System.Windows.Threading;
using System.Linq;
using Microsoft.Win32;

namespace ComputeFileHashes
{
    /// <summary>
    /// Handles the common user interface aspects of ComputeFileHashes.
    /// </summary>
    class ComputeFileHashesUI
    {
        // Collection of all files.
        private readonly ObservableCollection<HashFileInfo> _files = new ObservableCollection<HashFileInfo>();
        // Collection of files waiting to be hashed.
        private readonly BlockingQueue<HashFileInfo> _pendingFiles = new BlockingQueue<HashFileInfo>();
        // Worker thread for performing hashing.
        private readonly Thread _worker;
        // True iff aborting
        private bool _abort;
        // Dispatcher to use when triggering PropertyChanged (Silverlight-only)
        private readonly Dispatcher _dispatcher;

        /// <summary>
        /// Begins initialization of a new instance.
        /// </summary>
        public ComputeFileHashesUI(Dispatcher dispatcher)
        {
            // Initialize variables
            _dispatcher = dispatcher;
            _worker = new Thread(WorkerThreadFunction);
        }

        /// <summary>
        /// Completes initialization of a new instance.
        /// </summary>
        /// <param name="dataGrid">DataGrid for file display.</param>
        /// <param name="versionBlock">TextBlock for version display.</param>
        [SuppressMessage("Microsoft.Globalization", "CA1303:Do not pass literals as localized parameters", MessageId = "System.Windows.Controls.TextBlock.set_Text(System.String)", Justification = "Not translating.")]
        public void Initialize(DataGrid dataGrid, TextBlock versionBlock)
        {
            // Initialize variables
            dataGrid.ItemsSource = _files;
            versionBlock.Text = ComputeFileHashesCore.Version;

            // Start worker thread
            _worker.Start();
        }

        /// <summary>
        /// Signal worker to stop processing.
        /// </summary>
        public void Stop()
        {
            _abort = true;
            _pendingFiles.UnblockingEnqueue(null);
        }

        /// <summary>
        /// Non-UI worker thread implementation.
        /// </summary>
        [SuppressMessage("Microsoft.Design", "CA1031:DoNotCatchGeneralExceptionTypes", Justification = "No way to know what could be thrown")]
        private void WorkerThreadFunction()
        {
            // Create an instance of the core object
            using (var computeFileHashesCore = new ComputeFileHashesCore())
            {
                while (true)
                {
                    // Get the next file to hash
                    var hashFileInfo = _pendingFiles.BlockingDequeue();

                    // Exit if done
                    if ((null == hashFileInfo) || _abort)
                    {
                        break;
                    }
                    try
                    {
                        // Process the file
                        computeFileHashesCore.ProcessFile(hashFileInfo.FileInfo, (percent) => { hashFileInfo.Progress = percent; return !_abort; });

                        // Store the results of the HashAlgorithms
                        hashFileInfo.CRC32 = ComputeFileHashesCore.ConvertValueToString(computeFileHashesCore.Hashes["CRC32"].Hash);
                        hashFileInfo.MD5 = ComputeFileHashesCore.ConvertValueToString(computeFileHashesCore.Hashes["MD5"].Hash);
                        hashFileInfo.SHA1 = ComputeFileHashesCore.ConvertValueToString(computeFileHashesCore.Hashes["SHA1"].Hash);
                        hashFileInfo.SHA256 = ComputeFileHashesCore.ConvertValueToString(computeFileHashesCore.Hashes["SHA256"].Hash);
#if !SILVERLIGHT
                        hashFileInfo.SHA384 = ComputeFileHashesCore.ConvertValueToString(computeFileHashesCore.Hashes["SHA384"].Hash);
                        hashFileInfo.SHA512= ComputeFileHashesCore.ConvertValueToString(computeFileHashesCore.Hashes["SHA512"].Hash);
                        hashFileInfo.RIPEMD160 = ComputeFileHashesCore.ConvertValueToString(computeFileHashesCore.Hashes["RIPEMD160"].Hash);
#endif
                    }
                    catch (Exception e)
                    {
                        // Indicate failure
                        hashFileInfo.CRC32 = hashFileInfo.MD5 = hashFileInfo.SHA1 = hashFileInfo.SHA256 = "×";
#if !SILVERLIGHT
                        hashFileInfo.SHA384 = hashFileInfo.SHA512 = hashFileInfo.RIPEMD160 = hashFileInfo.CRC32;
#endif
                        hashFileInfo.Details = e.Message;
                    }
                }
            }
        }

        /// <summary>
        /// Show the OpenFileDialog for the user to select files.
        /// </summary>
        [SuppressMessage("Microsoft.Performance", "CA1811:AvoidUncalledPrivateCode", Justification = "Called from XAML-based event handler.")]
        public void ShowOpenFileDialog()
        {
            // Initialize the dialog
            var dialog = new OpenFileDialog
            {
                Multiselect = true,
#if !SILVERLIGHT
                ShowReadOnly = false,
#endif
            };

            // Show the dialog
            if (dialog.ShowDialog().Value)
            {
                // Queue a HashFileInfo for each selected file
#if SILVERLIGHT
                foreach(var file in dialog.Files)
#else
                foreach (var file in dialog.FileNames.Select(fn => new FileInfo(fn)))
#endif
                {
                    QueueFile(new HashFileInfo(file, _dispatcher));
                }
            }
        }

        /// <summary>
        /// Queue a file for processing.
        /// </summary>
        /// <param name="hashFileInfo">The file to process.</param>
        public void QueueFile(HashFileInfo hashFileInfo)
        {
            // Add the file to the complete list and queue it for processing
            _files.Add(hashFileInfo);
            _pendingFiles.UnblockingEnqueue(hashFileInfo);
        }
    }
}
