﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Security.Cryptography;
using System.Text;
using ComputeFileHashes;
using Microsoft.VisualStudio.TestTools.UnitTesting;

namespace HashAlgorithmTests
{
    public abstract class HashAlgorithmTests
    {
        protected abstract HashAlgorithm NewHashAlgorithm();
        protected abstract IEnumerable<TestCase> TestCases { get;}

        protected void CheckProperties(int hashSize)
        {
            var hashAlgorithm = NewHashAlgorithm();
            Assert.AreEqual(true, hashAlgorithm.CanReuseTransform);
            Assert.AreEqual(true, hashAlgorithm.CanTransformMultipleBlocks);
            Assert.AreEqual(hashSize, hashAlgorithm.HashSize);
            Assert.AreEqual(1, hashAlgorithm.InputBlockSize);
            Assert.AreEqual(1, hashAlgorithm.OutputBlockSize);
        }

        protected void CheckTestCasesUsingComputeHash()
        {
            foreach (var testCase in TestCases)
            {
                CheckComputeHash(testCase.Buffer, testCase.HashString, true);
                CheckComputeHash(testCase.Buffer, testCase.HashString, false);
            }
        }

        protected void CheckTestCasesUsingTransformBlock()
        {
            foreach (var testCase in TestCases)
            {
                // Note: blockSize is deliberately larger than the buffer to exercise TransformFinalBlock
                for (var blockSize = 1; blockSize <= testCase.Buffer.Length + 1; blockSize++)
                {
                    CheckTransformBlock(testCase.Buffer, testCase.HashString, true, blockSize, false);
                    CheckTransformBlock(testCase.Buffer, testCase.HashString, false, blockSize, false);
                    CheckTransformBlock(testCase.Buffer, testCase.HashString, false, blockSize, true);
                }
            }
        }

        protected void CheckComputeHash(byte[] buffer, string hashString, bool initialize)
        {
            var hashAlgorithm = NewHashAlgorithm();
            if (initialize)
            {
                hashAlgorithm.Initialize();
            }
            var hash = hashAlgorithm.ComputeHash(buffer);
            Assert.AreEqual(hashString, ComputeFileHashesCore.ConvertValueToString(hash));
        }

        protected void CheckTransformBlock(byte[] buffer, string hashString, bool initialize, int blockSize, bool emptyFinalBlock)
        {
            var hashAlgorithm = NewHashAlgorithm();
            if (initialize)
            {
                hashAlgorithm.Initialize();
            }
            var index = 0;
            while (index + blockSize <= buffer.Length)
            {
                hashAlgorithm.TransformBlock(buffer, index, blockSize, null, 0);
                index += blockSize;
            }
            if (emptyFinalBlock)
            {
                hashAlgorithm.TransformBlock(buffer, index, buffer.Length - index, null, 0);
                index += buffer.Length - index;
            }
            hashAlgorithm.TransformFinalBlock(buffer, index, buffer.Length - index);
            Assert.AreEqual(hashString, ComputeFileHashesCore.ConvertValueToString(hashAlgorithm.Hash));
        }

        protected void CheckComputeHashReuse()
        {
            var hashAlgorithm = NewHashAlgorithm();
            var testCase = TestCases.Last();
            Assert.AreEqual(testCase.HashString, ComputeFileHashesCore.ConvertValueToString(hashAlgorithm.ComputeHash(testCase.Buffer)));
            Assert.AreEqual(testCase.HashString, ComputeFileHashesCore.ConvertValueToString(hashAlgorithm.ComputeHash(testCase.Buffer)));
        }

        protected void CheckTransformBlockReuse()
        {
            var hashAlgorithm = NewHashAlgorithm();
            var testCase = TestCases.Last();
            hashAlgorithm.TransformBlock(testCase.Buffer, 0, testCase.Buffer.Length, null, 0);
            hashAlgorithm.TransformFinalBlock(testCase.Buffer, 0, 0);
            Assert.AreEqual(testCase.HashString, ComputeFileHashesCore.ConvertValueToString(hashAlgorithm.Hash));
            hashAlgorithm.Initialize();
            hashAlgorithm.TransformBlock(testCase.Buffer, 0, testCase.Buffer.Length, null, 0);
            hashAlgorithm.TransformFinalBlock(testCase.Buffer, 0, 0);
            Assert.AreEqual(testCase.HashString, ComputeFileHashesCore.ConvertValueToString(hashAlgorithm.Hash));
        }

        protected void CheckHashAccessAfterReInitialize()
        {
            var hashAlgorithm = NewHashAlgorithm();
            var testCase = TestCases.Last();
            Assert.AreEqual(testCase.HashString, ComputeFileHashesCore.ConvertValueToString(hashAlgorithm.ComputeHash(testCase.Buffer)));
            hashAlgorithm.Initialize();
            var hash = hashAlgorithm.Hash;
        }

        protected void CheckInvalidHashAccess()
        {
            var hashAlgorithm = NewHashAlgorithm();
            var hash = hashAlgorithm.Hash;
        }

        protected void CheckMisingTransformFinalBlock()
        {
            var hashAlgorithm = NewHashAlgorithm();
            var testCase = TestCases.Last();
            hashAlgorithm.TransformBlock(testCase.Buffer, 0, testCase.Buffer.Length, null, 0);
            Assert.AreEqual(testCase.HashString, ComputeFileHashesCore.ConvertValueToString(hashAlgorithm.Hash));
        }

        protected void CheckInvalidTransformBlockReuse()
        {
            var hashAlgorithm = NewHashAlgorithm();
            var testCase = TestCases.Last();
            hashAlgorithm.TransformBlock(testCase.Buffer, 0, testCase.Buffer.Length, null, 0);
            hashAlgorithm.TransformFinalBlock(testCase.Buffer, 0, 0);
            Assert.AreEqual(testCase.HashString, ComputeFileHashesCore.ConvertValueToString(hashAlgorithm.Hash));
            hashAlgorithm.TransformBlock(testCase.Buffer, 0, testCase.Buffer.Length, null, 0);
        }

        protected byte[] GetConstantRandomBytes(int count)
        {
            // Use count as seed so results are always different - but repeatable
            var rand = new Random(count);
            var buffer = new byte[count];
            rand.NextBytes(buffer);
            return buffer;
        }
    }

    public class TestCase
    {
        public byte[] Buffer { get; private set; }
        public string HashString { get; private set; }
        public TestCase(byte[] buffer, string hashString)
        {
            Buffer = buffer;
            HashString = hashString;
        }
    }

    [TestClass]
    public class CRC32Tests : HashAlgorithmTests
    {
        protected override HashAlgorithm NewHashAlgorithm()
        {
            return new Delay.CRC32();
        }

        protected override IEnumerable<TestCase> TestCases
        {
            get
            {
                yield return new TestCase(Encoding.UTF8.GetBytes(""), "00000000");
                yield return new TestCase(Encoding.UTF8.GetBytes("1"), "83DCEFB7");
                yield return new TestCase(Encoding.UTF8.GetBytes("to"), "D787D2C4");
                yield return new TestCase(Encoding.UTF8.GetBytes("tre"), "8C68973D");
                yield return new TestCase(Encoding.UTF8.GetBytes("four"), "90C1667D");
                yield return new TestCase(Encoding.UTF8.GetBytes("cinco"), "8F0AE08E");
                yield return new TestCase(Encoding.UTF8.GetBytes("sixsix"), "BA8BBC6A");
                yield return new TestCase(Encoding.UTF8.GetBytes("septyni"), "916A902F");
                yield return new TestCase(Encoding.UTF8.GetBytes("otteotte"), "F3992050");
                yield return new TestCase(GetConstantRandomBytes(256), "4C969589");
                yield return new TestCase(GetConstantRandomBytes(1024), "8B0B425B");
            }
        }

        [TestMethod]
        public void TestProperties()
        {
            CheckProperties(32);
        }

        [TestMethod]
        public void TestCasesUsingComputeHash()
        {
            CheckTestCasesUsingComputeHash();
        }

        [TestMethod]
        public void TestCasesUsingTransformBlock()
        {
            CheckTestCasesUsingTransformBlock();
        }

        [TestMethod]
        public void TestComputeHashReuse()
        {
            CheckComputeHashReuse();
        }

        [TestMethod]
        public void TestTransformBlockReuse()
        {
            CheckTransformBlockReuse();
        }

        [TestMethod]
        [ExpectedException(typeof(NullReferenceException))]
        public void TestHashAccessAfterReInitialize()
        {
            CheckHashAccessAfterReInitialize();
        }

        [TestMethod]
        [ExpectedException(typeof(NullReferenceException))]
        public void TestInvalidHashAccess()
        {
            CheckInvalidHashAccess();
        }

        [TestMethod]
        [ExpectedException(typeof(CryptographicException))]
        public void TestMisingTransformFinalBlock()
        {
            CheckMisingTransformFinalBlock();
        }

        [TestMethod]
        [ExpectedException(typeof(CryptographicException))]
        public void TestInvalidTransformBlockReuse()
        {
            CheckInvalidTransformBlockReuse();
        }
    }

    public class CRC32Subclass : Delay.CRC32
    {
        public void CallHashCoreWithNullArray()
        {
            HashCore(null, 0, 0);
        }
    }

    [TestClass]
    public class CRC32SubclassTests
    {
        [TestMethod]
        [ExpectedException(typeof(ArgumentNullException))]
        public void TestHashCoreWithNullArray()
        {
            using (var hashAlgorithm = new CRC32Subclass())
            {
                hashAlgorithm.CallHashCoreWithNullArray();
            }
        }
    }

    public abstract class MD5HashAlgorithmTests : HashAlgorithmTests
    {
        protected override IEnumerable<TestCase> TestCases
        {
            get
            {
                yield return new TestCase(Encoding.UTF8.GetBytes(""), "D41D8CD98F00B204E9800998ECF8427E");
                yield return new TestCase(Encoding.UTF8.GetBytes("1"), "C4CA4238A0B923820DCC509A6F75849B");
                yield return new TestCase(Encoding.UTF8.GetBytes("to"), "01B6E20344B68835C5ED1DDEDF20D531");
                yield return new TestCase(Encoding.UTF8.GetBytes("tre"), "7E764AA6B7529530855F0373606D1886");
                yield return new TestCase(Encoding.UTF8.GetBytes("four"), "8CBAD96ACED40B3838DD9F07F6EF5772");
                yield return new TestCase(Encoding.UTF8.GetBytes("cinco"), "64E38E70EFB84F926568D0522A105C33");
                yield return new TestCase(Encoding.UTF8.GetBytes("sixsix"), "54D0B65DBCE1BCAE13E1329438D021BF");
                yield return new TestCase(Encoding.UTF8.GetBytes("septyni"), "A823A4F92B21BD53A4F469AC42721119");
                yield return new TestCase(Encoding.UTF8.GetBytes("otteotte"), "1E294C3021C0E48FABD18E7F21896572");
                yield return new TestCase(GetConstantRandomBytes(60), "206E90C85A388E021D2E64682721D939");
                yield return new TestCase(GetConstantRandomBytes(256), "8B1A1BDC3E533D7FEE6A5983E411C075");
                yield return new TestCase(GetConstantRandomBytes(1024), "6B6C8AC360B6A9FE97A15EFE32B87997");
            }
        }

        protected void CheckProperties()
        {
            CheckProperties(128);
        }

        protected void CheckLargeBufferTestCase()
        {
            var buffer = new byte[1 << 29]; // Specific size threshold for MD5 algorithm
            var hashString = "AA559B4E3523A6C931F08F4DF52D58F2";
            CheckTransformBlock(buffer, hashString, true, 1024, false);
        }
    }

    [TestClass]
    public class MD5ManagedTests : MD5HashAlgorithmTests
    {
        protected override HashAlgorithm NewHashAlgorithm()
        {
            return new Delay.MD5Managed();
        }

        [TestMethod]
        public void TestProperties()
        {
            CheckProperties();
        }

        [TestMethod]
        public void TestCasesUsingComputeHash()
        {
            CheckTestCasesUsingComputeHash();
        }

        [TestMethod]
        public void TestCasesUsingTransformBlock()
        {
            CheckTestCasesUsingTransformBlock();
        }

        [TestMethod]
        public void CheckLargeBuffer()
        {
            CheckLargeBufferTestCase();
        }

        [TestMethod]
        public void TestComputeHashReuse()
        {
            CheckComputeHashReuse();
        }

        [TestMethod]
        public void TestTransformBlockReuse()
        {
            CheckTransformBlockReuse();
        }

        [TestMethod]
        [ExpectedException(typeof(NullReferenceException))]
        public void TestHashAccessAfterReInitialize()
        {
            CheckHashAccessAfterReInitialize();
        }

        [TestMethod]
        [ExpectedException(typeof(NullReferenceException))]
        public void TestInvalidHashAccess()
        {
            CheckInvalidHashAccess();
        }

        [TestMethod]
        [ExpectedException(typeof(CryptographicException))]
        public void TestMisingTransformFinalBlock()
        {
            CheckMisingTransformFinalBlock();
        }

        [TestMethod]
        [ExpectedException(typeof(CryptographicException))]
        public void TestInvalidTransformBlockReuse()
        {
            CheckInvalidTransformBlockReuse();
        }
    }

    [TestClass]
    public class MD5CryptoServiceProviderTests : MD5HashAlgorithmTests
    {
        protected override HashAlgorithm NewHashAlgorithm()
        {
            return new System.Security.Cryptography.MD5CryptoServiceProvider();
        }

        [TestMethod]
        public void TestProperties()
        {
            CheckProperties();
        }

        [TestMethod]
        public void TestCasesUsingComputeHash()
        {
            CheckTestCasesUsingComputeHash();
        }

        [TestMethod]
        public void TestCasesUsingTransformBlock()
        {
            CheckTestCasesUsingTransformBlock();
        }

        [TestMethod]
        public void TestLargeBuffer()
        {
            CheckLargeBufferTestCase();
        }

        [TestMethod]
        public void TestComputeHashReuse()
        {
            CheckComputeHashReuse();
        }

        [TestMethod]
        public void TestTransformBlockReuse()
        {
            CheckTransformBlockReuse();
        }

        [TestMethod]
        //[ExpectedException(typeof(NullReferenceException))] // Lack of an exception here seems like a bug in the .NET implementation
        public void TestHashAccessAfterReInitialize()
        {
            CheckHashAccessAfterReInitialize();
        }

        [TestMethod]
        [ExpectedException(typeof(NullReferenceException))]
        public void TestInvalidHashAccess()
        {
            CheckInvalidHashAccess();
        }

        [TestMethod]
        [ExpectedException(typeof(CryptographicUnexpectedOperationException))]
        public void TestMisingTransformFinalBlock()
        {
            CheckMisingTransformFinalBlock();
        }

        [TestMethod]
        [ExpectedException(typeof(CryptographicException))]
        public void TestInvalidTransformBlockReuse()
        {
            CheckInvalidTransformBlockReuse();
        }
    }

    public class MD5ManagedSubclass : Delay.MD5Managed
    {
        public void CallHashCoreWithNullArray()
        {
            HashCore(null, 0, 0);
        }
    }

    [TestClass]
    public class MD5ManagedSubclassTests
    {
        [TestMethod]
        [ExpectedException(typeof(ArgumentNullException))]
        public void TestHashCoreWithNullArray()
        {
            using (var hashAlgorithm = new MD5ManagedSubclass())
            {
                hashAlgorithm.CallHashCoreWithNullArray();
            }
        }
    }
}
