﻿using System;
using System.ComponentModel;
using System.Diagnostics.CodeAnalysis;
using System.Globalization;
using System.IO;
using System.Security;
using System.Windows.Threading;

namespace ComputeFileHashes
{
    /// <summary>
    /// Represents a file to be hashed.
    /// </summary>
    public class HashFileInfo : INotifyPropertyChanged
    {
        /// <summary>
        /// The file information.
        /// </summary>
        public FileInfo FileInfo { get; private set; }

        /// <summary>
        /// The full name of the file.
        /// </summary>
        public string FullName { get; private set; }

        /// <summary>
        /// The short name of the file.
        /// </summary>
        public string Name { get; private set; }

        /// <summary>
        /// The progress of the hashing operation ([0..1]).
        /// </summary>
        public double Progress
        {
            get
            {
                return _progress;
            }
            set
            {
                _progress = value;
                FirePropertyChanged("Progress");
                FirePropertyChanged("FormattedProgress");
            }
        }
        private double _progress;

        /// <summary>
        /// The formatted progress of the hashing operation (X.Y%).
        /// </summary>
        public string FormattedProgress
        {
            get
            {
                return string.Format(CultureInfo.CurrentCulture, "{0:0.0}%", Progress);
            }
        }

        /// <summary>
        /// The CRC32 hash of the file.
        /// </summary>
        [SuppressMessage("Microsoft.Naming", "CA1709:IdentifiersShouldBeCasedCorrectly", MessageId = "CRC", Justification = "Casing is consistent with acronym")]
        public string CRC32
        {
            get
            {
                return _crc32;
            }
            set
            {
                _crc32 = value;
                FirePropertyChanged("CRC32");
            }
        }
        private string _crc32;

        /// <summary>
        /// The MD5 hash of the file.
        /// </summary>
        public string MD5
        {
            get
            {
                return _md5;
            }
            set
            {
                _md5 = value;
                FirePropertyChanged("MD5");
            }
        }
        private string _md5;

        /// <summary>
        /// The SHA1 hash of the file.
        /// </summary>
        [SuppressMessage("Microsoft.Naming", "CA1709:IdentifiersShouldBeCasedCorrectly", MessageId = "SHA", Justification = "Casing is consistent with acronym")]
        public string SHA1
        {
            get
            {
                return _sha1;
            }
            set
            {
                _sha1 = value;
                FirePropertyChanged("SHA1");
            }
        }
        private string _sha1;

        /// <summary>
        /// The SHA256 hash of the file.
        /// </summary>
        [SuppressMessage("Microsoft.Naming", "CA1709:IdentifiersShouldBeCasedCorrectly", MessageId = "SHA", Justification = "Casing is consistent with acronym")]
        public string SHA256
        {
            get
            {
                return _sha256;
            }
            set
            {
                _sha256 = value;
                FirePropertyChanged("SHA256");
            }
        }
        private string _sha256;

#if !SILVERLIGHT
        /// <summary>
        /// The SHA384 hash of the file.
        /// </summary>
        [SuppressMessage("Microsoft.Naming", "CA1709:IdentifiersShouldBeCasedCorrectly", MessageId = "SHA", Justification = "Casing is consistent with acronym")]
        public string SHA384
        {
            get
            {
                return _sha384;
            }
            set
            {
                _sha384 = value;
                FirePropertyChanged("SHA384");
            }
        }
        private string _sha384;

        /// <summary>
        /// The SHA512 hash of the file.
        /// </summary>
        [SuppressMessage("Microsoft.Naming", "CA1709:IdentifiersShouldBeCasedCorrectly", MessageId = "SHA", Justification = "Casing is consistent with acronym")]
        public string SHA512
        {
            get
            {
                return _sha512;
            }
            set
            {
                _sha512 = value;
                FirePropertyChanged("SHA512");
            }
        }
        private string _sha512;

        /// <summary>
        /// The RIPEMD160 hash of the file.
        /// </summary>
        [SuppressMessage("Microsoft.Naming", "CA1709:IdentifiersShouldBeCasedCorrectly", MessageId = "RIPEMD", Justification = "Matching algorithm name and .NET use.")]
        public string RIPEMD160
        {
            get
            {
                return _ripemd160;
            }
            set
            {
                _ripemd160 = value;
                FirePropertyChanged("RIPEMD160");
            }
        }
        private string _ripemd160;
#endif

        /// <summary>
        /// Any details about the hash process (ex: error information).
        /// </summary>
        public string Details
        {
            get
            {
                return _details;
            }
            set
            {
                _details = value;
                FirePropertyChanged("Details");
            }
        }
        private string _details;

        // Dispatcher to use when triggering PropertyChanged (Silverlight-only)
        [SuppressMessage("Microsoft.Performance", "CA1823:AvoidUnusedPrivateFields", Justification = "Needed for Silverlight; present on WPF to avoid extra #if/#endif blocks")]
        private readonly Dispatcher _dispatcher;

        /// <summary>
        /// Initializes a new instance.
        /// </summary>
        /// <param name="fileInfo">File being hashed.</param>
        /// <param name="dispatcher">Dispatcher to use.</param>
        public HashFileInfo(FileInfo fileInfo, Dispatcher dispatcher)
        {
            // Validate arguments
            if (null == fileInfo)
            {
                throw new ArgumentNullException("fileInfo");
            }
            if (null == dispatcher)
            {
                throw new ArgumentNullException("dispatcher");
            }

            // Initialize variables
            FileInfo = fileInfo;
            _dispatcher = dispatcher;
            try
            {
                FullName = fileInfo.FullName;
            }
            catch (SecurityException)
            {
                // Unable to access full path
                FullName = fileInfo.Name;
            }
            Name = fileInfo.Name;
            CRC32 = MD5 = SHA1 = SHA256 = "-";
#if !SILVERLIGHT
            SHA384 = SHA512 = RIPEMD160 = CRC32;
#endif
        }

        /// <summary>
        /// Fires the PropertyChanged event.
        /// </summary>
        /// <param name="propertyName">Name of property that changed.</param>
        private void FirePropertyChanged(string propertyName)
        {
            var handler = PropertyChanged;
            if (null != handler)
            {
#if SILVERLIGHT
                // Silverlight needs this to happen on the UI thread
                _dispatcher.BeginInvoke(() => handler.Invoke(this, new PropertyChangedEventArgs(propertyName)));
#else
                // WPF allows it on any thread
                handler.Invoke(this, new PropertyChangedEventArgs(propertyName));
#endif
            }
        }

        /// <summary>
        /// INotifyPropertyChanged.PropertyChanged event.
        /// </summary>
        public event PropertyChangedEventHandler PropertyChanged;
    }
}
